package com.acme.client;

import java.util.Iterator;
import java.util.List;

import com.google.gwt.core.client.EntryPoint;
import com.google.gwt.core.client.GWT;
import com.google.gwt.user.client.Command;
import com.google.gwt.user.client.DeferredCommand;
import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.rpc.ServiceDefTarget;
import com.google.gwt.user.client.ui.Button;
import com.google.gwt.user.client.ui.ClickListener;
import com.google.gwt.user.client.ui.Grid;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Hyperlink;
import com.google.gwt.user.client.ui.KeyboardListener;
import com.google.gwt.user.client.ui.KeyboardListenerAdapter;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.Panel;
import com.google.gwt.user.client.ui.PopupPanel;
import com.google.gwt.user.client.ui.RootPanel;
import com.google.gwt.user.client.ui.TextBox;
import com.google.gwt.user.client.ui.VerticalPanel;
import com.google.gwt.user.client.ui.Widget;
import com.google.gwt.user.client.ui.Window;

/**
 * A CRUD address book. You can Create, Read, Update and 
 * Delete contacts from a database. This application uses 
 * remote procedure calls (RPCs) that use Hibernate on the 
 * server.
 */
public class AddressBookApplication implements EntryPoint {   
  // Create an address book service and a grid. We will 
  // populate the grid with contacts obtained from the service.
  private final AddressBookServiceAsync ws = 
   (AddressBookServiceAsync)GWT.create
      (AddressBookService.class);  
  private final ServiceDefTarget target = 
      (ServiceDefTarget) ws;
  private final Grid grid = new Grid();
    
  /*
   * This is the entry point for the address book application.
   * 
   * @see com.google.gwt.core.client.EntryPoint#onModuleLoad()
   */
  public void onModuleLoad() {
    // Set the entry point for the address book service. That
    // entry point must match the remote servlet's path (see
    // com.acme.public.AddressBookApplication.gwt.xml for the
    // servlet definition, or /WEB-INF/web.xml in web mode).
     target.setServiceEntryPoint(GWT.getModuleBaseURL() 
                                + "/addressBookService");
     
     // Load the address book from the database and create
     // the corresponding user interface.
     loadAddressBook();
     createUI();
  }
  
  /*
   * Use the address book service to get the list of 
   * contacts from the database, and populate the grid 
   * accordingly.
   */
  private void loadAddressBook() {
   ws.getContacts(new AsyncCallback() { // invoke RPC
     /* GWT calls onSuccess() if the preceeding RPC 
      * succeeded, and passes the result of that call, 
      * which is a list
      * of com.acme.client.Trip objects.
      */ 
      public void onSuccess(Object result) {
          // Clear the grid and populate it with the result of 
         // the RPC.
         grid.clear();
         
         List contacts = (List)result;
         resizeGrid(contacts);
         repopulateGrid(contacts);
      }
      
      /*
       * GWT calls onFailure() if the preceeding RPC failed,
       * and passes a throwable. 
       */
      public void onFailure(Throwable t) {
          Window.alert("Remote procedure call failed: "
                       + t.getMessage());
      }

      /*
       * Resize the grid based on the current contacts in 
       * the database.
       */
      private void resizeGrid(List contacts) {      
         if (contacts.size() == 0) {
            grid.resize(1, 1);
            grid.setText(0, 0, "No contacts in the database");
            return;
         }
         
         // We have some contacts, (otherwise we would've 
         // returned from this method) so we resize the 
         // grid accordingly.
         grid.resize(contacts.size(), 3);
      }
      
      /*
       * Repopulate the grid with the current contacts.
       */
      private void repopulateGrid(List contacts) {
         Iterator it = contacts.iterator();
         
         // Iterate over the list of contacts and populate the
         // grid with contact properties name and email.
         for (int row=0; it.hasNext(); ++row) {
             final Contact contact = (Contact)it.next();
             
           // Each row in the grid has an edit link, email, and
           // delete button.
            final Button button = new Button("delete");
            final Hyperlink link = 
                new Hyperlink(contact.getName(), 
                              contact.getName());
            
            // Add the link, text, and button to grid row
            grid.setWidget(row, 0, link); // edit hyperlink
            grid.setText(row, 1, contact.getEmail()); // email
            grid.setWidget(row, 2, button); // delete button
            
            // Add click listeners to the edit link and 
            // delete button.
            link.addClickListener(new ClickListener() {
               public void onClick(Widget sender) {
                  editContact(sender, contact);
               }
            });
            
            button.addClickListener(new ClickListener() {
               public void onClick(Widget sender) {
                  deleteContact(contact);
               }
            });
         }
      }
   });
  }
  
  /*
   * This method creates the user interface by populating the
   * root panel with a horizontal panel that contains the 
   * <i>Your contacts:</i> label and a vertical panel. The
   * vertical panel contains the grid and the
   *  <i>New contact...</i> button.
   */
  private void createUI() {
    // Create the New contact... button and 
   // add a click listener to it.
    final Button newContactButton = 
        new Button("New contact...");   

    newContactButton.addClickListener(new ClickListener() {
      public void onClick(Widget sender) {
         createContact(sender);
      }             
    });
     
    // Create the vertical panel, add the grid and 
    // New contact... button, and set the grid's style.
    VerticalPanel vp = new VerticalPanel();
    vp.add(grid);
    vp.add(newContactButton);
    grid.addStyleName("contactsTable");
    
    // Create the horizontal panel, add the 
    //Your contacts: label,
    // and add the vertical panel to the horizontal panel.
    final HorizontalPanel hp = new HorizontalPanel();
    final Label headingLabel = new Label("Your contacts:");
    hp.add(headingLabel);
    hp.add(vp);
    
    // Clear the root panel of all of its widgets and add the
    // horizontal panel.
    RootPanel.get().clear();
    RootPanel.get().add(hp);
  }
  
  /*
   * This callback is used by saveContact(), 
   * updateContact(), and deleteContact().
   */
  private final AsyncCallback asyncCallback = 
      new AsyncCallback() {
   public void onSuccess(Object result) {
      loadAddressBook();
   }
   public void onFailure(Throwable t) {
      Window.alert("RPC call failed: " + t.getMessage());
   }      
  };
  
  /*
   * This method is called by the delete button's 
   * click listener.
   */
  private void deleteContact(Contact contact) {
     ws.deleteContact(contact, asyncCallback);     
  }
  
  /*
   * This method is called by a click listener attached to
   * the contact panel popup's button when editing a contact.
   */  
  private void updateContact(Contact contact) {
     ws.updateContact(contact, asyncCallback);
  }
  
  /*
   * This method is called by a click listener attached to
   * the contact panel popup's button when creating a contact.
   */    
  private void saveContact(Contact contact) {
     ws.saveContact(contact, asyncCallback);
  }
  
  /*
   * This method is called from the New contact... button's 
   * click listener. It creates a popup panel, and sets the 
   * poup's style and position. Then it adds a panel, created by 
   * createContactPanel(), to the popup panel.
   */
  private void createContact(Widget sender) {
    // The true argument to the PopupPanel constructor is an
    // auto-hide boolean. When true, poup panels close 
    // when the user clicks outside of them.
     final PopupPanel popup = new PopupPanel(true);
     popup.addStyleName("createContactPopup");
     popup.setPopupPosition(sender.getAbsoluteLeft() + 10,
                            sender.getAbsoluteTop() + 10);
     
     // Add the popup's contents and show the popup.
     popup.add(createContactPanel(popup, null, 
                                  "Create a contact"));
     popup.show();
  }
  
  /*
   * This method is called by the edit link's click listener 
   * It creates a popup panel, populates it with widgets, and
   * shows the popup panel.
   */  
  private void editContact(Widget sender, Contact contact) {
     PopupPanel popup = new PopupPanel(true);
     popup.addStyleName("createContactPopup");
     popup.setPopupPosition(sender.getAbsoluteLeft() + 10,
                          sender.getAbsoluteTop() + 10);
     Panel p = createContactPanel(popup, contact, 
                                  "Edit Contact");
     popup.add(p);
     popup.show();     
  }
  
  /*
   * Creates a panel that's used to edit or create a contact.
   * This method is called by the createContact() and 
   * editContact() methods (which are in turn called by 
   * event listeners). Those methods create a popup panel 
   * that they pass to this method. That popup panel's 
   * hide method is called in an event listener
   * implemented in this method.
   */
  private Panel createContactPanel(final PopupPanel popup,
                                   final Contact contact,
                                   final String buttonText) {
    // Create the grid with 5 rows and 2 columns.
     final Grid grid = new Grid(5, 2);
     
     // Create the widgets that go in the grid.
     final Label heading = new Label(buttonText);
     heading.addStyleName("heading");
     
     final Label namePrompt = new Label("Name: ");
     final TextBox nameTextBox = new TextBox();

     final Label emailPrompt = new Label("Email: ");
     final TextBox emailTextBox = new TextBox();
     
     final Button button = new Button(buttonText);
     
     // If we're editing (as opposed to creating) a contact,
     // populate the name and email textboxes accordingly.
     if (contact != null) {
        nameTextBox.setText(contact.getName());
        emailTextBox.setText(contact.getEmail());
     }

     // Define a single keyboard listener used by both the 
     // name and email textboxes.
     final KeyboardListener fieldListener = 
        new KeyboardListenerAdapter() {
       public void onKeyPress(Widget sender, 
                         char keycode, int modifiers) {
          if (keycode == KeyboardListener.KEY_TAB) {
             if (sender == nameTextBox) {
                emailTextBox.setFocus(true);
             }
             else
                nameTextBox.setFocus(true);
          }
          else if (keycode == KeyboardListener.KEY_ENTER) {
             button.click();
          }
       }
     };
     
     // Add the preceeding keyboard listener to the name and
     // email textboxes.
     nameTextBox.addKeyboardListener(fieldListener);
     emailTextBox.addKeyboardListener(fieldListener);
     
     // Add a click listener to the button that either 
     // creates a new contact or edits the current contact, 
     // depending upon whether we're editing a contact or 
     // creating a new one.
     button.addClickListener(new ClickListener() {
      public void onClick(Widget sender) {
        // Get the name and email entered in the 
       // corresponding textboxes.
         String name = nameTextBox.getText();
         String email = emailTextBox.getText();
         
         // If we're editing (not creating), then update the
         // contact's information; otherwise, create a new 
         // contact
         if (contact != null) {
            contact.setName(name);
            contact.setEmail(email);
            updateContact(contact);
         }
         else // Creating a new contact
            saveContact(new Contact(null, name, email));
         
         // Hide the popup panel.
         popup.hide();
      }
     });
     
     // We can't give the name textbox focus now because it 
     // hasn't been displayed yet. So we use a deferred 
     // command to give the name textbox focus after the 
     // current stack of events has been handled. By that 
     // time, the name textbox will be displayed.
     DeferredCommand.add(new Command() {
        public void execute() {
           nameTextBox.setFocus(true);
        }
     });
          
     // Add widgets to the grid
     grid.setWidget(0, 0, namePrompt);     
     grid.setWidget(0, 1, nameTextBox);     
     grid.setWidget(1, 0, emailPrompt);     
     grid.setWidget(1, 1, emailTextBox);
     
     // Create a vertical panel and add the heading, 
     // grid, and button.
     final VerticalPanel vp = new VerticalPanel();     
     vp.add(heading);
     vp.add(grid);
     vp.add(button);
     
     return vp;
  }
}
