package com.acme.client.dnd;

import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.ui.AbsolutePanel;
import com.google.gwt.user.client.ui.MouseListenerAdapter;
import com.google.gwt.user.client.ui.Widget;

/*
 * A mouse listener that drags widgets. Subclasses must 
 * implement getNextLeft() and getNextTop(), signifying 
 * the next position for a widget, given a proposed position.
 * 
 * On mouse down, this listener records the x and y offsets
 * of the mouse click from the upper left corner of the widget. 
 * It then calls {@link}DOM.setCapture() so the widget 
 * being dragged has exclusive access to mouse events 
 * until the mouse is released.
 * 
 * As the mouse is being dragged, this listener calculates a 
 * proposed new position based on the current location of the 
 * mouse. That algorithm, calculating the next position based 
 * on mouse position, is pluggable. Most of the time, you will 
 * want the widget being dragged to simply follow the mouse, 
 * but sometimes, you might want to restrain the widget to move 
 * in the horizontal or vertical directions, for instance. 
 * This class's abstract methods encapsulate that algorithm.
 * 
 * When the mouse is released, this listener calls {@link}
 * DOM.releaseCapture(), thereby making mouse events available 
 * to everyone again.
 * 
 */
public abstract class AbstractMouseDragger 
                extends MouseListenerAdapter
                implements MouseDragger {
   private boolean dragging = false;
   private int xoffset, yoffset;
   
   protected abstract int getNextLeft(int proposedLeft, 
                                 Widget draggable);
                     
   protected abstract int getNextTop(int proposedTop, 
                                 Widget draggable);
   
   public void onMouseDown(Widget draggable, int x, int y) {
      xoffset = x;
      yoffset = y;
      DOM.setCapture(draggable.getElement());
      dragging = true;
   }
   public void onMouseMove(Widget draggable, int x, int y) {
        if (dragging) {
           // All draggable widgets must have an AbsolutePanel
           // for a parent.
           AbsolutePanel ap = 
               (AbsolutePanel)draggable.getParent();
           int proposedx = x 
                           + ap.getWidgetLeft(draggable) 
                           - xoffset;
           int proposedy = y 
                           + ap.getWidgetTop(draggable) 
                           - yoffset;
           
           int nextx = getNextLeft(proposedx, draggable);
           int nexty = getNextTop(proposedy, draggable);
           
           ap.setWidgetPosition(draggable, nextx, nexty);
        }
    }
   public void onMouseUp(Widget draggable, int x, int y) {
      dragging = false;
      DOM.releaseCapture(draggable.getElement());
   }
   
   protected int checkLeftBounds(int proposedLeft, 
                               Widget widget) {
      AbsolutePanel panel = (AbsolutePanel)widget.getParent();
      int widgetWidth = widget.getOffsetWidth();
      int panelWidth = panel.getOffsetWidth();
      int nextx = proposedLeft;
      
      if (proposedLeft + widgetWidth > panelWidth)
      nextx = panelWidth - widgetWidth;
      
      nextx = nextx < 0 ? 0 : nextx;
      return nextx;
   }
   protected int checkTopBounds(int proposedTop, 
                             Widget widget) { 
      AbsolutePanel panel = (AbsolutePanel)widget.getParent();
      int widgetHeight = 
            widget.getOffsetHeight();               
      int panelHeight = panel.getOffsetHeight();
      int nexty = proposedTop;
      
      if (proposedTop + widgetHeight > panelHeight)
      nexty = panelHeight - widgetHeight;
      
      nexty = nexty < 0 ? 0 : nexty;
      return nexty;
   }
}
